#!/usr/bin/perl
#
# PPP Remote Control: A Perl CGI script acting as remote control for pppd.
# Copyright (C) 2000 by Josef Drexler <jdrexler@julian.uwo.ca>
# Distributable under the terms of the GNU General Public License. For more
# information see the file COPYING which should have come with this file.
#
# Usage:
# ------
# Simply install in your cgi-bin directory and try it, possibly
# after changing the next few lines;
# ppp-rc is able to start any pppd connection listed in /etc/ppp/peers,
# list the active connections or stop those that it started (i.e. those
# run by the same user your web server is running under).

	# pppd should detach after establishing a network protocol and call
	# the selected ISP, with the options from /etc/ppp/peers/%s.
$pppd = "/usr/sbin/pppd updetach call %s";
	# You may need to use the following if pppd isn't world executable
	# but can be run from e.g. the "dialout" group. The web server's 
	# UID must be a member of the group and it can't have a password.
	# See gpasswd(1) for more information.
#$pppd = "/usr/bin/sg dialout -c \'/usr/sbin/pppd updetach call %s\'";
$peerlist = "/etc/ppp/peers/*";
$ifconfig = "/sbin/ifconfig";
$pidfile = "/var/run/%s.pid";
$killsignal = 15;	# 1 for HUP, 15 for TERM
$procnetdev = "/proc/net/dev";
$successurl = "/";
#$successurl = "http://www.yahoo.com/";
$showbytes = 1;
$callmethod = 1;

# Description of the options:
# $pppd is the command line to be run for a new connection, where %s will be
#       the connection name
# $peerlist is a list of files that are to be listed as possible connections.
#       Must be acceptable as argument to "ls -1".
# $ifconfig is the path for the ifconfig executable
# $pidfile is where pppd stores its pid file, with %s being the interface name
# $killfile is the signal used to kill pppd processes. Use TERM (15) because
#	HUP (1) will restart connections with the "persist" or "demand" option 
#	and KILL (9) won't allow for a clean exit.
# $procnetdev points to the file where device stat are available, for 
#	showing bytes RX/TX statistics. Only with 2.2.x kernel
# $showbytes can be set to 1 if $procnetdev is set and you have a 2.2.x kernel
#	If 0 it will show the packet counts instead.
# $successurl is where clicking the "OK" button of a successful connection
#	will take you.
# $callmethod selects whether pppd should be run in the background (though not
#	detached) to show output as soon as it arrives (1) or if ppp-rc should
#	wait for pppd to return (0).
#
#
# General idea about how it operates:
#
# ppp-rc implements both the GET and POST methods.  GET is used for the
# initial control page, and can also display a status-only page (call it
# with .../ppp-rc.cgi?status=1). POST is used to start and stop connections.
# This makes it easier to give different access authorities for each method
# using the web server's capabilities. A sample .htaccess file for the
# Apache server has been included.


# Divine the used method, parse the form data if any, and
# output the appropriate page
&init;
if($ENV{'REQUEST_METHOD'} eq "GET"){
	# GETting is either Control, Status or Redirect
	DecodeFormData($ENV{'QUERY_STRING'});
	if ($formdata{'action'} eq "redirect"){
		&Redirect;
	}else{
		&printHead;
		if($formdata{'status'}){
			&StatusPage;
		}else{
			&ControlPage;
		}
	}
}elsif($ENV{'REQUEST_METHOD'} eq "POST"){
	# POSTing can be starting or stopping (or redirect)
	$| = 1;		# Set autoflush
	read(STDIN, $buffer, $ENV{'CONTENT_LENGTH'});
	DecodeFormData($buffer);
	if ($formdata{'action'} eq "redirect"){
		&Redirect;
	}else{
		&printHead;
		if ($formdata{'action'} eq "start"){
			&StartPage;
		}elsif ($formdata{'action'} eq "stop"){
			&StopPage;
		}else{
			&ErrorPage("Invalid action \"$formdata{'action'}\".\n");
		}
	}
}else{
	&printHead;
	&ErrorPage("Invalid method \"$ENV{'REQUEST_METHOD'}\".\n");
}
&printFoot;

exit(0);
#
#
# Sends a 301 Moved output and redirects to the selected page
#
sub Redirect{
	print "Location: $formdata{'RedirectURL'}\n";
	print "Status: 301 Moved Permanently\n";
	print "Content-Type: text/html\n\n";
	print "<HTML>\n<HEAD>\n\t<TITLE>301 Moved Permanently</TITLE>\n";
	print "</HEAD>\n<BODY>\n";
	print "<H1>Moved Permanently</H1>\n";
	print "The document has moved <a href=\"$formdata{'RedirectURL'}\">";
	print "here</a>.<p>\n";
}
#
# Gives a list of active connections for  the separate status-only page
# (URL .../ppp-rc.cgi?status=1)
#
sub StatusPage{
	&comment("StatusPage");
	print "<H1>The following connections are active:</H1>\n";
	&ShowActiveConnections;
	$referrer = $ENV{'HTTP_REFERER'};
	if (! $referrer){
		$referrer=$ENV{'SCRIPT_NAME'};
	}
	print "<br>\n<form action=\"$referrer\" method=\"get\">\n";
	&KeepOptions("\t");
	print "\t<input type=submit value=Back>\n";
	print "</form>\n";
	&comment("/StatusPage");
}
#
# The control page to start, list and stop connections. Displayed if a
# GET with no options was requested.
#
sub ControlPage{
	&comment("ControlPage");
	print "<H1>Remote Control options:</H1>\n";
	print "<H2>Start new connection</H2>\n";
	&ShowStartConnections;
	print "<H2>Connection information</H2>\n";
	&ShowActiveConnections;
	print "<hr width=\"100%\">\n<form method=\"get\">\n";
	&KeepOptions("\t");
	print "\t<input type=submit value=\"Refresh\">\n</form>\n";
	&comment("/ControlPage");
}
#
# Called when the "new connection" button is pressed
#
sub StartPage{
	&comment("StartPage");
	&InitPPPExitText;
	$ISP=$formdata{'ISP'};
	if ( !$ISP || ($ISP eq "Select Connection")){
		print "<H1>Error</H1>\n<P>Please select a connection.\n";
		print "<form method=\"get\">\n\t<input type=submit value=\"Back\">\n";
		&KeepOptions("\t");
		print "</form>\n";
		return;
	}
	print "<H1>Starting Connection \"$ISP\"</H1>\n";
	$command = sprintf($pppd, $ISP);
	print "<P>Executing \'$command\'.\n<P>";
	$transcriptstart = "Session transcript:<BR>\n<BLOCKQUOTE><PRE>";
	$transcriptend = "</PRE></BLOCKQUOTE>\n<P>";
	if ($callmethod){
		$pid = open(PPPDOUTPUT, "-|");
		unless (defined $pid) {
			print "An error has occured while trying to fork.\n";
			return;
		}
		if ($pid){	# parent (remains ppp-rc)
			while(<PPPDOUTPUT>){
				print "$transcriptstart$_";
				$transcriptstart="";
			}
			close(PPPDOUTPUT);
			if (!$transcriptstart){
				print "$transcriptend";
			}
		}else{		# child to-be-pppd
			{ exec "/bin/sh", "-c", $command  };
			print "Error running $command, not found?\n";
			exit;
		}
		# waitpid($pid,0);	# WHY don't I need this??
	}else{
	
		$result = `$command`;
		if ($result){
			print "$transcriptstart$result$transcriptend";
		}
	}
	$exitcode = $? >> 8;
	print "The exit code was $exitcode, which means ";
	print "\"$exittext[$exitcode]\".\n";
	print "<P>\n<table>\n";
	print "\t<tr>\t<td><form method=\"post\">\n";
	print "\t\t\t<input type=hidden value=\"$successurl\" name=\"RedirectURL\">\n";
	print "\t\t\t<input type=hidden value=\"redirect\" name=\"action\">\n";
	&KeepOptions("\t\t\t");
	print "\t\t\t<input type=submit value=\"OK\">\n";
	print "\t\t</form></td>\n";
	print "\t\t<td><form method=\"post\">\n";
	print "\t\t\t<input type=hidden value=\"$ISP\" name=\"ISP\">\n";
	print "\t\t\t<input type=hidden value=\"start\" name=\"action\">\n";
	&KeepOptions("\t\t\t");
	print "\t\t\t<input type=submit value=\"Retry\">\n";
	print "\t\t</form></td>\n";	
	print "\t\t<td><form method=\"get\">\n"; 
	&KeepOptions("\t\t\t");
	print "\t\t\t<input type=submit value=\"Back\">\n";
	print "\t\t</form></td>\n";
	print "\t</tr>\n</table>\n";
	&comment("/StartPage");
}
#
# Called to stop the selected connection
#
sub StopPage{
	&comment("StopPage");
	$iface=$formdata{'interface'};
	if (!$iface){
		print "<H1>Error</H1>\n<P>Please select a connection.\n";
		print "<form method=\"get\">\n\t<input type=submit value=\"Back\">\n";
		print "</form>\n";
		return;
	}
	print "<H1>Stopping connection $iface</H1>\n";
	$filename = sprintf($pidfile,$iface);
	open(PIDFILE, $filename);
	$pid = 0 + <PIDFILE>;		# read pid and force it to be integer
	close(PIDFILE);
	print "<P>Killing $iface daemon, PID $pid:\n";
	$result = kill $killsignal, $pid;
	if ($result){
		print "OK!\n";
	}else{
		print "Failed! (Wrong user?)\n";
	}
	print "<p>\n<form method=\"get\">\n";
	&KeepOptions("\t");
	print "\t<input type=submit value=\"Back\">\n</form>\n";
	&comment("/StopPage");
}
#
# Displays an error message
#
sub ErrorPage{
	&comment("ErrorPage");
	print "<H1>An error has occured!</H1>\n";
	print "@_";
	&comment("/ErrorPage");
}
#
# Displays the form from which a connection can be started, by listing
# the entries in $peerlist.
#
sub ShowStartConnections{
	&comment("ShowStartConnections");
	print "<form method=\"post\">\n";
	@ispentries = `ls -1 $peerlist`;
	print "\t<select name=\"ISP\">\n";
	if ($#ispentries){
		print "\t\t<option>Select Connection</option>\n";
	}
	foreach $isp (@ispentries){
		$isp =~ s/.*\/([^\/]+)/$1/;
		chomp $isp;
		print "\t\t<option";
		if ($formdata{'selectstart'} eq $isp){
			print " selected";
		}
		print ">$isp</option>\n";
	}
	print "\t</select>\n";
	print "\t<input type=submit value=\"New Connection\">\n";
	&KeepOptions("\t");
	print "\t<input type=hidden name=action value=start>\n";
	print "</form>\n";
	&comment("/ShowStartConnections");
}
#
# Shows a table of active connections, and the "Stop connection" button
#
sub ShowActiveConnections{
	&comment("ShowActiveConnections");
	&GetPPPInterfaces;
	$count=0;
	print "<form method=\"post\">\n\t<table border>\n";
	print "\t\t<tr>\t<th>Interface</th><th>Local IP</th>";
	print "<th>Remote IP</th>\n\t\t\t";
	if ($showbytes){
		print "<th>Received Bytes</th><th>Sent Bytes</th>"
	}else{
		print "<th>Received Packets</th><th>Sent Packets</th>"
	}
	print "\n\t\t\t<th>Activated on</th><th>Time active</th></tr>\n";
	for ($ifnum=0;$ifnum<=$#interface;$ifnum++){
		if ($interface[$ifnum] =~ /ppp/){
			&GetInterfaceProperties($interfacedata[$ifnum]);
			&GetExtendedInterfaceProperties($interface[$ifnum]);
			&ShowInterface;
			$count++;
		}
	}
	if (!$count){
		print "\t\t<tr><td>No PPP interfaces active</td></tr>\n";
	}
	print "\t</table>\n\t<br>\n";
	&KeepOptions("\t");
	print "\t<input type=hidden name=action value=stop>\n";
	print "\t<input type=submit value=\"Stop the selected connection\">\n";
	print "</form>\n";
	&comment("/ShowActiveConnections");
}
#
# Show one row with one interface for the connection table
#
sub ShowInterface{
	print "\t\t<tr>\t<td>";
	print "<input type=\"radio\" name=\"interface\" ";
	if ($formdata{'selectstop'} eq $interface[$ifnum]){
		print "checked ";
	}
	print "value=\"$interface[$ifnum]\">";
	print "$interface[$ifnum]</td>\n";
	print "\t\t\t<td>$ifaceprop{'inet addr'}</td>";
	print "<td>$ifaceprop{'P-t-P'}</td>\n\t\t\t";
	if ($showbytes){
		print "<td>$ifaceprop{'RX bytes'}</td><td>$ifaceprop{'TX bytes'}</td>\n";
	}else{
		print "<td>$ifaceprop{'RX packets'}</td><td>$ifaceprop{'TX packets'}</td>\n";
	}
	print "\t\t\t<td>$ifaceprop{'PID ctime local'}</td><td>$ifaceprop{'time active'}</td></tr>\n";
}
#
# Defines hidden inputs in a form to keep options between GETs and POSTs
#
sub KeepOptions{
	foreach $option (@keepoptions){
		if ($formdata{$option}){
			print "@_<input type=hidden name=\"$option\" value=\"$formdata{$option}\">\n";
		}
	}
}
#
# Collect all options from the GET or POST and put the in $formdata
#
sub DecodeFormData{
	# Split the name-value pairs
	@pairs = split(/&/, $_[0]);

	foreach $pair (@pairs){
		($name, $value) = split(/=/, $pair);
		$value =~ tr/+/ /;
		$value =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;
		$name =~ tr/+/ /;
		$name =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg; 

		$formdata{$name} = $value;
	}
}
#
# Enumerates the PPP interfaces
#
sub GetPPPInterfaces{
	@ifconfigresult = `$ifconfig`;

	$ifline = "";
	$ifnum=0;
	$#interface = -1;
	$#interfacedata = -1;
	$#interfaceactive = -1;
	foreach $line (@ifconfigresult){
		if ($line eq "\n"){
			$ifline =~ tr/\\\n/*/;
			@iface = split(/ /, $ifline, 2);
			$interface[$ifnum]=$iface[0];
			$interfacedata[$ifnum] = "$iface[1]";
			$interfaceactive{$iface[0]} = 1;
			$ifline = ""; $ifnum++;
		}else{
			$ifline = $ifline . $line;
		}
	}
}
#
# Get active time and bytes transferred for a selected interface
#
sub GetExtendedInterfaceProperties{
	$ifaceprop{"RX bytes"} = "n/a";
	$ifaceprop{"TX bytes"} = "n/a";
	$iface = $_[0];
	open(PROCINFO, $procnetdev);
	while (<PROCINFO>) {
		if ($_ !~ /\|/){
			$_ =~ tr/:/ /;
			@entries = split;
			if ($entries[0] eq $iface) {
				if ($#entries == 16){	# 2.2.x kernel only!
					$ifaceprop{"RX bytes"} = $entries[1];
					$ifaceprop{"TX bytes"} = $entries[9];
				}
			}
		}
	}
	close(PROCINFO);
	$filename = sprintf($pidfile,$iface);
	@pidstat = stat($filename);
	$ifaceprop{"PID ctime"} = $pidstat[10];
	$ifaceprop{"PID ctime GM"} = gmtime($pidstat[10]);
	$ifaceprop{"PID ctime local"} = localtime($pidstat[10]);
	$active = time - $pidstat[10];
	$activetext = "";
	use integer;
	$value = $active / (24*3600);
	$active -= $value * 24*3600;
	if ($value || $activetext){
		$activetext=$activetext . $value . "d";
	}
	$value = $active / 3600;
	$active -= $value * 3600;
	if ($value || $activetext){
		$activetext=$activetext . $value . "h";
	}
	$value = $active / 60;
	$active -= $value * 60;
	if ($value || $activetext){
		$activetext=$activetext . $value . "m";
	}
	$value = $active;
	$active -= $value;
	if ($value || $activetext){
		$activetext=$activetext . $value . "s";
	}
	$ifaceprop{"time active"} = $activetext;
}
#
# Parse the ifconfig output
#
sub GetInterfaceProperties{
	@pairs = split(/ /, $_[0]);

	$lastsingle = "";
	foreach $pair (@pairs){
		($name, $value) = split(/:/, $pair);
		if (!$value) {
			$lastsingle = $name;
		}else{
			if ($lastsingle){
				$name = $lastsingle . " " . $name;
				$lastsingle = "";
			}
			$ifaceprop{$name} = $value;
		}
	}
}
#
# Text message for each of the pppd exit codes.
#
sub InitPPPExitText{
	$exittext[0]="Connection successfully established";
	$exittext[1]="Fatal error";
	$exittext[2]="Error in given options";
	$exittext[3]="Pppd not suid root nor started by root";
	$exittext[4]="Kernel does not support PPP";
	$exittext[5]="Terminated by SIGINT, SIGTERM or SIGHUP";
	$exittext[6]="The device could not be locked";
	$exittext[7]="The device could not be openend";
	$exittext[8]="The connect script failed";
	$exittext[9]="The pty command could not be run";
	$exittext[10]="PPP negotiation failed";
	$exittext[11]="The peer failed or refused to authenticate itself";
	$exittext[12]="The link was terminated because it was idle";
	$exittext[13]="The link was terminated because the connect time limit was reached";
	$exittext[14]="Callback was negotiated and an incoming call should arrive shortly";
	$exittext[15]="The link was terminated because the peer was not responding to echo requests";
	$exittext[16]="The link was terminated by the modem hanging up";
	$exittext[17]="PPP negotiation failed because serial loopback was detected";
	$exittext[18]="The init script failed";
	$exittext[19]="We failed to authenticate ourselves (check the passwords)";
}
#
# General page header
#
sub printHead{
	print "Content-Type: text/html\n\n";
	print "<HTML>\n<HEAD>\n\t<TITLE>PPP Remote Control</TITLE>\n";
	print "</HEAD>\n<BODY>\n";
}
#
# General page footer
#
sub printFoot{
	print "</BODY>\n</HTML>\n";
}
#
# Print a comment to show which sub the output came from
#
sub comment{
	print "<!-- @_ --->\n";
}
#
# Initialization
#
sub init{
	$keepoptions[0] = "selectstart";
	$keepoptions[1] = "selectstop";
}